@@ -1,4 +1,6 @@
 """
+[BSD 2-CLAUSE LICENSE]
+
 Copyright Alex Yu 2021
 
 Redistribution and use in source and binary forms, with or without
@@ -24,6 +26,7 @@ ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 """
 
+import os.path as osp
 import torch
 import numpy as np
 from torch import nn
@@ -38,7 +41,6 @@ try:
 except:
     _C = None
 
-torch.autograd.set_detect_anomaly(True)
 
 class _SVEQueryVerticalFunction(torch.autograd.Function):
     @staticmethod
@@ -63,6 +65,68 @@ class _SVEQueryVerticalFunction(torch.autograd.Function):
 
         return grad_data, None, None, None, None, None
 
+class _N3TreeView:
+    def __init__(self, tree, key):
+        self.tree = tree
+        self.key = key
+        self._value = None;
+
+    def __repr__(self):
+        return "N3TreeView(" + repr(self.value()) + ")"
+
+    def value(self):
+        if self._value is None:
+            self._value = self.tree.values().__getitem__(self.key)
+        return self._value
+
+    def __torch_function__(self, func, types, args=(), kwargs=None):
+        if kwargs is None:
+            kwargs = {}
+        new_args = []
+        for arg in args:
+            if isinstance(arg, _N3TreeView):
+                new_args.append(arg.value())
+            else:
+                new_args.append(arg)
+        return func(*new_args, **kwargs)
+
+    def set(self, value):
+        self.tree._push_to_leaf()
+        leaf_node = self.tree._all_leaves()
+        key1 = self.key[0] if isinstance(self.key, tuple) else self.key
+        leaf_node = leaf_node.__getitem__(key1)
+        if isinstance(self.key, tuple):
+            leaf_node_sel = (*leaf_node.T, *self.key[1:])
+        else:
+            leaf_node_sel = (*leaf_node.T,)
+        self.tree.data.data[leaf_node_sel] = value
+
+    @property
+    def shape(self):
+        return self.value().shape
+
+    @property
+    def requires_grad(self):
+        return self.value().requires_grad
+
+    @requires_grad.setter
+    def requires_grad(self, value):
+        self.value().requires_grad = value
+
+# Redirect functions to Tensor
+def _redirect_funcs():
+    redir_funcs = ['__add__', '__mul__', '__sub__', '__floordiv__',
+                   '__mod__', '__div__', '__radd__', '__rsub__', '__rmul__',
+                   '__rdiv__', 'item', 'size', 'dim', 'detach', 'cpu', 'cuda', 'to',
+                   'float', 'double', 'int', 'long']
+    def redirect_func(redir_func):
+        def redir_impl(self, *args, **kwargs):
+            val = self.value()
+            return getattr(self.value(), redir_func)(*args, **kwargs)
+        setattr(_N3TreeView, redir_func, redir_impl)
+    for redir_func in redir_funcs:
+        redirect_func(redir_func)
+_redirect_funcs()
 
 class N3Tree(nn.Module):
     """
@@ -70,7 +134,7 @@ class N3Tree(nn.Module):
     """
     def __init__(self, N=4, data_dim=4, depth_limit=4,
             init_reserve=4, init_refine=0, geom_resize_fact=1.5, padding_mode="zeros",
-            radius=0.5, center=[0.5, 0.5, 0.5]):
+            radius=0.5, center=[0.5, 0.5, 0.5], map_location="cpu"):
         """
         :param N branching factor N
         :param data_dim size of data stored at each leaf
@@ -95,17 +159,17 @@ class N3Tree(nn.Module):
                 init_reserve += (N ** i) ** 3
 
         self.register_parameter("data", nn.Parameter(
-            torch.zeros(init_reserve, N, N, N, data_dim)))
+            torch.zeros(init_reserve, N, N, N, data_dim, device=map_location)))
         self.register_buffer("child", torch.zeros(
-            init_reserve, N, N, N, dtype=torch.int32))
+            init_reserve, N, N, N, dtype=torch.int32, device=map_location))
         self.register_buffer("parent_depth", torch.zeros(
-            init_reserve, 2, dtype=torch.int32))
+            init_reserve, 2, dtype=torch.int32, device=map_location))
 
-        self.register_buffer("_n_internal", torch.tensor(1))
-        self.register_buffer("max_depth", torch.tensor(0))
+        self.register_buffer("_n_internal", torch.tensor(1, device=map_location))
+        self.register_buffer("max_depth", torch.tensor(0, device=map_location))
 
-        radius = torch.tensor(radius)
-        center = torch.tensor(center)
+        radius = torch.tensor(radius, device=map_location)
+        center = torch.tensor(center, device=map_location)
         self.register_buffer("invradius", 0.5 / radius, persistent=False)
         self.register_buffer("offset", 0.5 * (1.0 - center / radius), persistent=False)
 
@@ -139,10 +203,11 @@ class N3Tree(nn.Module):
             indices = self._transform_coord(indices)
 
             if self.padding_mode == "zeros":
-                outside_mask = ((indices >= 1.0) | (indices < 1.0)).any(dim=-1)
-                indices = indices[outside_mask]
+                outside_mask = ((indices >= 1.0) | (indices < 0.0)).any(dim=-1)
+                indices = indices[~outside_mask]
 
             n_queries, _ = indices.shape
+            indices.clamp_(0.0, 1.0 - 1e-10)
             ind = indices.clone()
 
             node_ids = torch.zeros(n_queries, dtype=torch.long, device=indices.device)
@@ -151,6 +216,7 @@ class N3Tree(nn.Module):
             remain_mask = torch.ones(n_queries, dtype=torch.bool, device=indices.device)
             while remain_mask.any():
                 ind_floor = torch.floor(ind[remain_mask] * self.N)
+                ind_floor.clamp_max_(self.N - 1)
                 sel = (node_ids[remain_mask], *(ind_floor.long().T),)
 
                 deltas = self.child[sel]
@@ -198,8 +264,8 @@ class N3Tree(nn.Module):
             indices = self._transform_coord(indices)
 
             if self.padding_mode == "zeros":
-                outside_mask = ((indices >= 1.0) | (indices < 1.0)).any(dim=-1)
-            indices.clamp_(0.0, 1.0)
+                outside_mask = ((indices >= 1.0) | (indices < 0.0)).any(dim=-1)
+            indices.clamp_(0.0, 1.0 - 1e-10)
 
             n_queries, _ = indices.shape
             ind = indices.clone()
@@ -207,8 +273,12 @@ class N3Tree(nn.Module):
             result = torch.zeros((n_queries, self.data_dim), dtype=torch.float32,
                                   device=indices.device)
             remain_mask = torch.ones(n_queries, dtype=torch.bool, device=indices.device)
+            if self.padding_mode == "zeros":
+                remain_mask &= ~outside_mask
+
             while remain_mask.any():
                 ind_floor = torch.floor(ind[remain_mask] * self.N)
+                ind_floor.clamp_max_(self.N - 1)
                 sel = (node_ids[remain_mask], *(ind_floor.long().T),)
 
                 deltas = self.child[sel]
@@ -405,7 +475,7 @@ class N3Tree(nn.Module):
         """
         return self.parent_depth.shape[0]
 
-    def values(self):
+    def values(self, depth=None):
         """
         Get a list of all leaf values in tree
         Side effect: pushes values to leaf.
@@ -414,29 +484,87 @@ class N3Tree(nn.Module):
         self._push_to_leaf()
         leaf_node = self._all_leaves()
         leaf_node_sel = (*leaf_node.T,)
-        return self.data[leaf_node_sel]
+        data = self.data[leaf_node_sel]
+        if depth is not None:
+            depths = self.parent_depth[leaf_node[:, 0], 1]
+            data = data[depths == depth]
+        return data
+
+    @property
+    def depth(self):
+        return self.depths()
 
     def depths(self):
         """
         Get a list of leaf depths in tree,
-        in same format as values().
+        in same order as values(), corners().
         Root is at depth 0.
-        :return (n_leaves, data_dim)
+        :return (n_leaves) int32
         """
         leaf_node = self._all_leaves()
         return self.parent_depth[leaf_node[:, 0], 1]
 
+    def corners(self, depth=None):
+        """
+        Get a list of leaf lower xyz corners in tree,
+        in same order as values(), depths().
+        :return (n_leaves, 3)
+        """
+        leaf_node = self._all_leaves()
+        corners = self._calc_corners(leaf_node)
+        if depth is not None:
+            depths = self.parent_depth[leaf_node[:, 0], 1]
+            corners = corners[depths == depth]
+        return corners
 
-    #  def render(self, rays, white_bkgd=True):
-    #      """
-    #      :param rays ray [origins (3), directions(3), near(1), far(1)] (Q, 8)
-    #      :return [rgb depth alpha]
-    #      """
-    #      # Extension required
-    #      return _C.render(self.data, self.child, rays,
-    #              self.offset,
-    #              self.invradius,
-    #              white_bkgd)
+    def savez(self, path):
+        data = {
+            "data_dim" : self.data_dim,
+            "child" : self.child.cpu(),
+            "parent_depth" : self.parent_depth.cpu(),
+            "n_internal" : self._n_internal.cpu().item(),
+            "max_depth" : self.max_depth.cpu().item(),
+            "invradius" : self.invradius.cpu().item(),
+            "offset" : self.offset.cpu(),
+            "depth_limit": self.depth_limit,
+            "geom_resize_fact": self.geom_resize_fact,
+            "padding_mode": self.padding_mode
+        }
+        if self.data_dim != 3 and self.data_dim != 4:
+            data["data"] = self.data.data.cpu()
+        else:
+            import imageio
+            data_path = osp.splitext(path)[0] + '_data.exr'
+            imageio.imwrite(data_path, self.data.data.cpu().reshape(-1,
+                self.N ** 2, self.data_dim))
+        np.savez_compressed(path, **data)
+
+    def loadz(self, path):
+        z = np.load(path)
+        device = self.data.data.device
+        self.data_dim = int(z["data_dim"])
+        self.child = torch.from_numpy(z["child"]).to(device)
+        self.N = self.child.shape[-1]
+        self.parent_depth = torch.from_numpy(z["parent_depth"]).to(device)
+        self._n_internal.fill_(z["n_internal"].item())
+        self.max_depth.fill_(z["max_depth"].item())
+        self.invradius.fill_(z["invradius"].item())
+        self.offset = torch.from_numpy(z["offset"]).to(device)
+        self.depth_limit = int(z["depth_limit"])
+        self.geom_resize_fact = float(z["geom_resize_fact"])
+        self.padding_mode = str(z["padding_mode"])
+        if _C is not None:
+            self.padding_mode_c = _C.parse_padding_mode(self.padding_mode)
+        if self.data_dim != 3 and self.data_dim != 4:
+            self.data.data = torch.from_numpy(z["data"]).to(device)
+        else:
+            import imageio
+            data_path = osp.splitext(path)[0] + '_data.exr'
+            self.data.data = torch.from_numpy(
+                        imageio.imread(data_path).reshape(
+                            -1, self.N, self.N, self.N, self.data_dim
+                        )
+                    ).to(device)
 
     # Magic
     def __repr__(self):
@@ -446,32 +574,21 @@ class N3Tree(nn.Module):
                     self.n_internal, self.capacity, self.max_depth.item())
 
     def __getitem__(self, key):
-        if isinstance(key, slice) and key.start is None and key.stop is None:
-            # Everything
-            return self
-        elif isinstance(key, tuple) and len(key) == 3:
+        if isinstance(key, tuple) and len(key) == 3:
             # Use x,y,z format
             return self.get(torch.tensor(key, dtype=torch.float32,
                 device=self.data.device)[None])[0]
         else:
-            raise NotImplementedError("Unsupported getitem magic")
+            return _N3TreeView(self, key)
 
     def __setitem__(self, key, val):
-        if isinstance(key, slice) and key.start is None and key.stop is None:
-            # Everything
-            self.data.data.zero_()
-            self.data.data[0] = self._make_val_tensor(val)
-        elif isinstance(key, int):
-            # By channel
-            self.data.data[..., key].zero_()
-            self.data.data[0, ..., key] = val
-        elif isinstance(key, tuple) and len(key) == 3:
+        if isinstance(key, tuple) and len(key) == 3:
             # Use x,y,z format
             key_tensor = torch.tensor(key, dtype=torch.float32,
                 device=self.data.device)[None]
             self.set(key_tensor, self._make_val_tensor(val))
         else:
-            raise NotImplementedError("Unsupported setitem magic")
+            _N3TreeView(self, key).set(val)
 
 
     def __iadd__(self, val):
@@ -514,6 +631,34 @@ class N3Tree(nn.Module):
         with_child_sel = (*with_child.T,)
         self.data.data[with_child_sel] = 0.0
 
+
+    def _calc_corners(self, nodes):
+        """
+        Compute lower bbox corners for given nodes
+        :nodes (Q, 4)
+        :return (Q, 3)
+        """
+        Q, _ = nodes.shape
+        filled = self.n_internal
+
+        curr = nodes.clone()
+        mask = torch.ones(Q, device=curr.device, dtype=torch.bool)
+        output = torch.zeros(Q, 3, device=curr.device, dtype=torch.float32)
+
+        while True:
+            output[mask] += curr[:, 1:]
+            output[mask] /= self.N
+
+            good_mask = curr[:, 0] != 0
+            if not good_mask.any():
+                break
+            mask[mask] = good_mask
+
+            curr = self._unpack_index(self.parent_depth[curr[good_mask, 0], 0].long())
+
+        return output
+
+
     def _pack_index(self, txyz):
         return txyz[:, 0] * (self.N ** 3) + txyz[:, 1] * (self.N ** 2) + \
                txyz[:, 2] * self.N + txyz[:, 3]
@@ -555,3 +700,16 @@ class N3Tree(nn.Module):
 
     def _transform_coord(self, indices):
         return torch.addcmul(self.offset, indices, self.invradius)
+
+    # Tensor analogy
+    @property
+    def shape(self):
+        return torch.Size((self.n_leaves, self.data_dim))
+
+    def dim(self):
+        return 2
+
+    def size(self, *args):
+        if len(args) == 1:
+            return self.shape[args[0]]
+        return self.shape
